from typing import Union

from taitale.utils.Parset import Parset
from taitale.utils.wrapper import parset_app


def imager(
    parset: Union[str, Parset, None] = None,
    **kwargs,
):
    if parset is None:
        parset = Parset(prefix="Cimager")
    elif isinstance(parset, str):
        parset_obj = Parset(prefix="Cimager")
        parset_obj.read_from_file(parset)
        parset = parset_obj
    elif not isinstance(parset, Parset):
        raise TypeError("parset must be either a string path or a Parset instance")

    imager_app = parset_app(
        name="imager",
        parset=parset,
        out_parset_name="taitale_cimager.in",
        cmd="imager -c {parset_name}",
        mpi_compartible=True,
    )

    imager_app(**kwargs)


imager.__doc__ = """

Perform imaging in a parallel/distributed environment or on a single computer system.

Parameters
----------
parset : str or Parset or None, optional
    Either a path to a parset file, a Parset instance, or None to create a new Parset.
    If None, creates a new empty Parset with prefix "Cimager".
**kwargs : dict
    Additional keyword arguments passed to the imager application.

Notes
-----
For further information on usage check the ASKAPsoft documentation for `imager <https://yandasoft.readthedocs.io/en/latest/calim/imager.html>`_

Example
-------
>>> from taitale.askap import imager
>>> # Using a parset file
>>> imager(
...    parset="cimager.in",
...     workers=4,
...     args = {
...         "dataset": "./1934-638_0.ms",
...         "Images.Names": "image.i.1934-638_0",
...         "Images.shape":[2048, 2048],
...         "Images.cellsize":['2arcsec', '2arcsec'],
...         "Images.image.1924-638.direction":"[19h39m25.0342, -63.42.45.623, J2000]",
...         "restore":"true",
...         "restore.beam":"fit",
...         "restore.beam.cutoff": "0.5"
...     }
... )

>>> # Using a Parset instance
>>> from taitale.utils.Parset import Parset
>>> parset = Parset(prefix="Cimager")
>>> parset.set("dataset", "./1934-638_0.ms")
>>> imager(parset=parset, workers=4)
"""



import argparse

from taitale.utils import CliArgs
from taitale.utils.wrapper import cli_app


def msconcat_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument("-o", "--output-file")

    parser.add_argument("positional", nargs="+")
    return parser


def msconcat(**kwargs):
    parser = msconcat_arguments()
    cli_args = CliArgs(parser=parser)
    msconcat_app = cli_app(
        name="msconcat",
        cmd="msconcat {option}",
        cli_args=cli_args,
    )

    msconcat_app(**kwargs)


msconcat.__doc__ = """
The msconcat program is used to concatenate multiple measurement sets

Parameters
----------
**kwargs : dict
    Additional keyword arguments passed to the msconcat application.

Notes
-----
For further information on usage check the ASKAPsoft documentation for `msconcat <https://yandasoft.readthedocs.io/en/latest/utils/msconcat.html>`_

Example
-------
>>> from taitale.askap import msconcat
>>> msconcat(args="-o output.ms input1.ms input2.ms input3.ms")
"""



from typing import Union

from taitale.utils.Parset import Parset
from taitale.utils.wrapper import parset_app


def linmos(
    parset: Union[str, Parset, None] = None,
    **kwargs,
):
    if parset is None:
        parset = Parset(prefix="linmos")
    elif isinstance(parset, str):
        parset_obj = Parset(prefix="linmos")
        parset_obj.read_from_file(parset)
        parset = parset_obj
    elif not isinstance(parset, Parset):
        raise TypeError("parset must be either a string path or a Parset instance")

    linmos_app = parset_app(
        name="linmos",
        parset=parset,
        out_parset_name="taitale_linmos.in",
        cmd="linmos -c {parset_name}",
        mpi_compartible=False,
    )

    linmos_app(**kwargs)


linmos.__doc__ = """
The linmos program is used to mosaic images together

Parameters
----------
parset : str or Parset or None, optional
    Either a path to a parset file, a Parset instance, or None to create a new Parset.
    If None, creates a new empty Parset with prefix "linmos".
**kwargs : dict
    Additional keyword arguments passed to the linmos application.

Notes
-----
For further information on usage check the ASKAPsoft documentation for `linmos <https://yandasoft.readthedocs.io/en/latest/calim/linmos.html>`_

Example
-------
>>> from taitale.askap import linmos
>>> # Using a parset file
>>> linmos(
...    parset="linmos.in",
...     args={
...         "names": "['image.0', 'image.1', 'image.2']",
...         "weights": "['weight.0', 'weight.1', 'weight.2']",
...         "outname": "mosaic",
...         "outweight": "weight",
...         "weighttype": "FromWeightImages",
...         "weightstate": "Inherent",
...         "nterms": "2",
...     }
... )

>>> # Using a Parset instance
>>> from taitale.utils.Parset import Parset
>>> parset = Parset(prefix="linmos")
>>> parset.set("names", "['image.0', 'image.1']")
>>> linmos(parset=parset)
"""



from typing import Union

from taitale.utils.Parset import Parset
from taitale.utils.wrapper import parset_app


def losoto(
    parset: Union[str, Parset, None] = None,
    **kwargs,
):
    if parset is None:
        parset = Parset()
    elif isinstance(parset, str):
        parset_obj = Parset()
        parset_obj.read_from_file(parset)
        parset = parset_obj
    elif not isinstance(parset, Parset):
        raise TypeError("parset must be either a string path or a Parset instance")

    # Pop h5parm from kwargs
    h5parm = kwargs.pop("h5parm", None)
    if h5parm is None:
        raise ValueError("h5parm parameter is required")

    losoto_app = parset_app(
        name="LoSoTo",
        parset=parset,
        out_parset_name="taitale_losoto.parset",
        cmd="losoto " + h5parm + " {parset_name}",
        mpi_compartible=False,
    )

    losoto_app(**kwargs)


losoto.__doc__ = """
Taitale interface to LoSoTo (LOFAR Solution Tool)

Parameters
----------
parset : str or Parset or None, optional
    Either a path to a parset file, a Parset instance, or None to create a new Parset.
    If None, creates a new empty Parset.
h5parm : str
    Path to the H5Parm file containing the solutions to process. This is a required parameter.
**kwargs : dict
    Additional keyword arguments passed to the LoSoTo application.

Notes
-----
For further information on usage check:
- LoSoTo GitHub repository: `<https://github.com/revoltek/losoto>`_
- LOFAR Imaging Cookbook: `<https://support.astron.nl/LOFARImagingCookbook/losoto.html>`_

Example
-------

>>> from taitale.tools import losoto
>>> # Using a parset file
>>> losoto(
...     parset="losoto.parset",
...     h5parm="solutions.h5",
...     args={
...         "soltab": "sol000/amplitude000",
...         "operation": "PLOT",
...         "axisInTable": "time",
...         "axisInCol": "ant",
...         "plotFlag": "True",
...         "prefix": "amp_",
...     },
... )

>>> # Using a Parset instance
>>> from taitale.utils.Parset import Parset
>>> parset = Parset()
>>> parset.set("soltab", "sol000/amplitude000")
>>> parset.set("operation", "SMOOTH")
>>> parset.set("axisToSmooth", "time")
>>> parset.set("windowSize", "3600")
>>> losoto(parset=parset, h5parm="solutions.h5")
"""



import argparse

from taitale.utils import CliArgs
from taitale.utils.wrapper import cli_app


def mslist_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument("--brief", action="store_true")
    parser.add_argument("--full", action="store_true")
    parser.add_argument("--what", action="store_true")
    parser.add_argument("--how", action="store_true")
    parser.add_argument("--tables", action="store_true")
    parser.add_argument("--data", action="store_true")
    parser.add_argument("--verbose", action="store_true")

    parser.add_argument("--datacolumn")
    parser.add_argument("--field")
    parser.add_argument("--spw")
    parser.add_argument("--antennna")
    parser.add_argument("--timerange")
    parser.add_argument("--correlation")
    parser.add_argument("--scan")
    parser.add_argument("--uvrange")
    parser.add_argument("--pagerow")
    parser.add_argument("--listfile")

    parser.add_argument("positional")
    return parser


def mslist(**kwargs):
    parser = mslist_arguments()
    cli_args = CliArgs(parser=parser)
    mslist_app = cli_app(
        name="mslist",
        cmd="mslist {option}",
        cli_args=cli_args,
        pipestderr=True,
    )

    mslist_app(**kwargs)


mslist.__doc__ = """
The mslist program is used to list the contents of a measurement set.

Parameters
----------
**kwargs : dict
    Additional keyword arguments passed to the mslist application.

Notes
-----
For further information on usage, check the ASKAPsoft documentation for `mslist <https://yandasoft.readthedocs.io/en/latest/calim/mslist.html>`_

Example
-------
>>> from taitale.askap import mslist
>>> mslist(args="--brief mydata.ms")
"""



from typing import Union

from taitale.utils.Parset import Parset
from taitale.utils.wrapper import parset_app


def ccalapply(
    parset: Union[str, Parset, None] = None,
    **kwargs,
):
    if parset is None:
        parset = Parset(prefix="Ccalapply")
    elif isinstance(parset, str):
        parset_obj = Parset(prefix="Ccalapply")
        parset_obj.read_from_file(parset)
        parset = parset_obj
    elif not isinstance(parset, Parset):
        raise TypeError("parset must be either a string path or a Parset instance")

    ccalapply_app = parset_app(
        name="ccalapply",
        parset=parset,
        out_parset_name="taitale_ccalapply.in",
        cmd="ccalapply -c {parset_name}",
        mpi_compartible=False,
    )

    ccalapply_app(**kwargs)


ccalapply.__doc__ = """
The purpose of ccalapply is to apply calibration parameters to Measurement Set

Parameters
----------
parset : str or Parset or None, optional
    Either a path to a parset file, a Parset instance, or None to create a new Parset.
    If None, creates a new empty Parset with prefix "Ccalapply".
**kwargs : dict
    Additional keyword arguments passed to the ccalapply application.

Notes
-----
For further information on usage check the ASKAPsoft documentation for `ccalapply <https://yandasoft.readthedocs.io/en/latest/calim/ccalapply.html>`_

Example
-------
>>> from taitale.askap import ccalapply
>>> # Using a parset file
>>> ccalapply(
...    parset="ccalapply.in",
...     args={
...         "dataset": "./1934-638_0.ms",
...         "calibaccess": "table",
...         "calibaccess.table": "1934-638_0.calib.bp",
...         "calibaccess.table.maxant": "36",
...         "calibrate.allowflag": "true"
...     }
... )

>>> # Using a Parset instance
>>> from taitale.utils.Parset import Parset
>>> parset = Parset(prefix="Ccalapply")
>>> parset.set("dataset", "./1934-638_0.ms")
>>> ccalapply(parset=parset)
"""



from typing import Union

from taitale.utils.Parset import Parset
from taitale.utils.wrapper import parset_app


def ccalibrator(
    parset: Union[str, Parset, None] = None,
    **kwargs,
):
    if parset is None:
        parset = Parset(prefix="Ccalibrator")
    elif isinstance(parset, str):
        parset_obj = Parset(prefix="Ccalibrator")
        parset_obj.read_from_file(parset)
        parset = parset_obj
    elif not isinstance(parset, Parset):
        raise TypeError("parset must be either a string path or a Parset instance")

    ccalibrator_app = parset_app(
        name="ccalibrator",
        parset=parset,
        out_parset_name="taitale_ccalibrator.in",
        cmd="ccalibrator -c {parset_name}",
        mpi_compartible=True,
    )

    ccalibrator_app(**kwargs)


ccalibrator.__doc__ = """
The ccalibrator program performs calibration in a parallel/distributed environment or on a single computer system.

Parameters
----------
parset : str or Parset or None, optional
    Either a path to a parset file, a Parset instance, or None to create a new Parset.
    If None, creates a new empty Parset with prefix "Ccalibrator".
**kwargs : dict
    Additional keyword arguments passed to the ccalibrator application.

Notes
-----
For further information on usage check the ASKAPsoft documentation for `ccalibrator <https://yandasoft.readthedocs.io/en/latest/calim/ccalibrator.html>`_

Example
-------
>>> from taitale.askap import ccalibrator
>>> # Using a parset file
>>> ccalibrator(
...    parset="ccalibrator.in",
...     args = {
...         "dataset": "1934-638_0.ms",
...         "nAnt": "36",
...         "nBeam": "1",
...         "solve": "antennagains",
...         "solver": "SVD",
...         "calibaccess": "table",
...         "calibaccess.table": "1934-638.calib",
...         "calibaccess.table.maxant": "36",
...         "sources.names": "['1934-638']",
...         "sources.1934-638.direction": "[19h39m25.027, -63.42.45.61, J2000]",
...         "sources.1934-638.model": "1934-638.model"
...     }
... )

>>> # Using a Parset instance
>>> from taitale.utils.Parset import Parset
>>> parset = Parset(prefix="Ccalibrator")
>>> parset.set("dataset", "1934-638_0.ms")
>>> ccalibrator(parset=parset, workers=4)
"""



from typing import Union

from taitale.utils.Parset import Parset
from taitale.utils.wrapper import parset_app


def cddcalibrator(
    parset: Union[str, Parset, None] = None,
    **kwargs,
):
    if parset is None:
        parset = Parset("Cddcalibrator")
    elif isinstance(parset, str):
        parset_obj = Parset("Cddcalibrator")
        parset_obj.read_from_file(parset)
        parset = parset_obj
    elif not isinstance(parset, Parset):
        raise TypeError("parset must be either a string path or a Parset instance")

    cddcalibrator_app = parset_app(
        name="Cddcalibrator",
        parset=parset,
        out_parset_name="taitale_cddcalibrator.in",
        cmd="Cddcalibrator -c {parset_name}",
        mpi_compartible=True,
    )

    cddcalibrator_app(**kwargs)


cddcalibrator.__doc__ = """
Taitale interface to Cddcalibrator

Parameters
----------
parset : str or Parset or None, optional
    Either a path to a parset file, a Parset instance, or None to create a new Parset.
    If None, creates a new empty Parset with prefix "Cddcalibrator".
**kwargs : dict
    Additional keyword arguments passed to the Cddcalibrator application.

Notes
-----
For further information on usage check the ASKAPsoft documentation for `Cddcalibrator <https://yandasoft.readthedocs.io/en/latest/calim/cddcalibrator.html>`_

Example
-------

>>> from taitale import taitale_env
>>> from taitale.askap import cddcalibrator
>>>
>>> env = taitale_env(runtime="container", image="csirocass/askapsoft")
>>>
>>> # Using direct arguments
>>> cddcalibrator(
...     env=env,
...     workers=6,  # Use 6 MPI workers
...     args={
...         "dataset": "observation.ms",
...         "nAnt": "36",
...         "nBeam": "36",
...         "refantenna": "1",
...         "calibaccess": "table",
...         "calibaccess.table": "ddcal.tab",
...     },
... )
>>>
>>> # Using a parset file
>>> cddcalibrator(
...     env=env,
...     workers=6,
...     parset="cddcalibrator.in",
... )
"""



import select
import subprocess
import time

import ipywidgets
from IPython.display import display


# TODO: This work only on Linux
def ICmdRunWithLogTrail(run_cmd, logfile):
    process = subprocess.Popen(f"{run_cmd}", shell=True)
    running = True

    out = ipywidgets.Output(
        layout={"border": "1px solid black", "width": "100%", "height": "360px"}
    )
    display(out)

    out.clear_output()

    tail = subprocess.Popen(
        ["tail", "-f", logfile], stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    tail_poll = select.poll()
    tail_poll.register(tail.stdout)

    chunks = 0
    while running:
        res = process.poll()
        running = res is None

        content = ""
        while tail_poll.poll(1):
            data = tail.stdout.readline()
            if data is not None:
                content += data.decode()
                chunks += 1

        if content != "":
            out.append_stdout(content)

        # 100 line buffer
        if chunks > 100:
            chunks = 0
            out.clear_output(wait=True)

        time.sleep(0.5)

    content = ""
    while tail_poll.poll(1):
        data = tail.stdout.readline()
        if data is not None:
            content += data.decode()
            chunks += 1

    if content != "":
        out.append_stdout(content)

    # Kill the tail process
    tail.kill()
    process.communicate()
    return process.returncode



from typing import Union

from taitale.utils.Parset import Parset
from taitale.utils.wrapper import parset_app


def ccontsubtract(
    parset: Union[str, Parset, None] = None,
    **kwargs,
):
    if parset is None:
        parset = Parset(prefix="CContSubtract")
    elif isinstance(parset, str):
        parset_obj = Parset(prefix="CContSubtract")
        parset_obj.read_from_file(parset)
        parset = parset_obj
    elif not isinstance(parset, Parset):
        raise TypeError("parset must be either a string path or a Parset instance")

    ccontsubtract_app = parset_app(
        name="ccontsubtract",
        parset=parset,
        out_parset_name="taitale_ccontsubtract.in",
        cmd="ccontsubtract -c {parset_name}",
        mpi_compartible=True,
    )

    ccontsubtract_app(**kwargs)


ccontsubtract.__doc__ = """
The purpose of ccontsubtract is to subtract continuum from a Measurement Set

Parameters
----------
parset : str or Parset or None, optional
    Either a path to a parset file, a Parset instance, or None to create a new Parset.
    If None, creates a new empty Parset with prefix "CContSubtract".
**kwargs : dict
    Additional keyword arguments passed to the ccontsubtract application.

Notes
-----
For further information on usage check the ASKAPsoft documentation for `ccontsubtract <https://yandasoft.readthedocs.io/en/latest/calim/ccontsubtract.html>`_

Example
-------
>>> from taitale.askap import ccontsubtract
>>> # Using a parset file
>>> ccontsubtract(
...    parset="ccontsubtract.in",
...     workers=2,
...     args={
...         "dataset": "./1934-638_0.ms",
...         "datacolumn": "DATA",
...         "outputcolumn": "CORRECTED_DATA",
...         "modelcolumn": "MODEL_DATA",
...         "nterms": "2",
...         "nchannel": "16416",
...         "spw": "0:0~16415",
...         "targetspec": "0:0~16415",
...         "freqframe": "topo",
...         "reffreq": "1.42GHz"
...     }
... )

>>> # Using a Parset instance
>>> from taitale.utils.Parset import Parset
>>> parset = Parset(prefix="CContSubtract")
>>> parset.set("dataset", "./1934-638_0.ms")
>>> ccontsubtract(parset=parset, workers=2)
"""



from typing import Union

from taitale.utils.Parset import Parset
from taitale.utils.wrapper import parset_app


def cbpcalibrator(
    parset: Union[str, Parset, None] = None,
    **kwargs,
):
    if parset is None:
        parset = Parset(prefix="Cbpcalibrator")
    elif isinstance(parset, str):
        parset_obj = Parset(prefix="Cbpcalibrator")
        parset_obj.read_from_file(parset)
        parset = parset_obj
    elif not isinstance(parset, Parset):
        raise TypeError("parset must be either a string path or a Parset instance")

    cbpcalibrator_app = parset_app(
        name="cbpcalibrator",
        parset=parset,
        out_parset_name="taitale_cbpcalibrator.in",
        cmd="cbpcalibrator -c {parset_name}",
        mpi_compartible=False,
    )

    cbpcalibrator_app(**kwargs)


cbpcalibrator.__doc__ = """
The cbpcalibrator program is a specialised tool for bandpass calibration.

Parameters
----------
parset : str or Parset or None, optional
    Either a path to a parset file, a Parset instance, or None to create a new Parset.
    If None, creates a new empty Parset with prefix "Cbpcalibrator".
**kwargs : dict
    Additional keyword arguments passed to the cbpcalibrator application.

Notes
-----
For further information on usage check the ASKAPsoft documentation for `cbpcalibrator <https://yandasoft.readthedocs.io/en/latest/calim/cbpcalibrator.html>`_

Example
-------
>>> from taitale.askap import cbpcalibrator
>>> # Using a parset file
>>> cbpcalibrator(
...     parset="cbpcalibrator.in",
...     args = {
...         "dataset": "./1934-638_0.ms",
...         "sources.names": "['src1']",
...         "sources.cal.flux.i": "1.0",
...         "sources.cal.direction.ra": "0.003",
...         "sources.cal.direction.dec": "0.0",
...         "sources.src1.direction": "[12h30m00.000, -45.00.00.000, J2000]",
...         "sources.src1.components": "['cal']",
...         "calibaccess": "table",
...         "nChan": "28",
...         "interval": "300.0s",
...         "calibaccess.table": "1934-638_0.calib.bp",
...         "refantenna": "4",
...         "calibaccess.table.maxant": "36"
...     }
... )

>>> # Using a Parset instance
>>> from taitale.utils.Parset import Parset
>>> parset = Parset(prefix="Cbpcalibrator")
>>> parset.set("dataset", "./1934-638_0.ms")
>>> cbpcalibrator(parset=parset)
"""



import configparser
from dataclasses import dataclass, field
from typing import Optional, Union

from .argumentset import ArgumentSet


# The concept of a group is to improve DX
@dataclass
class Parset(ArgumentSet):
    # @parameter prefix Defines the prefix of the Parset eg Csimulator, Cimager
    # @parameter group Defines the group values belong to
    prefix: Optional[str] = None
    group_name: Optional[str] = None
    prefix_ser: str = field(init=False)

    def __post_init__(self):
        super().__init__()
        self.prefix_ser = f"{self.prefix}." if self.prefix else ""

    # TODO: Find a better name for this
    def value(self, value):
        self.set(f"{self.prefix}", value)

    def comment(self, comment: str):
        # The simplest way to serialize define this, definetly can be improved
        comment = f"# {comment}"
        self.set(comment, comment)

    def merge(self, other: Union[ArgumentSet, "Parset"], dropPrefix=False):
        """Merge another ArgumentSet or Parset into this one.

        Args:
            other: The ArgumentSet or Parset to merge from
            dropPrefix: Only used when merging from another Parset. If True, drops the prefix when merging.
        """
        if isinstance(other, ArgumentSet) and not isinstance(other, Parset):
            # Simple merge for ArgumentSet
            for k, v in other.args.items():
                self.set(k, v)
        else:
            # Merge from another Parset
            if dropPrefix:
                for k, v in other.args.items():
                    if self.prefix and k.startswith(f"{self.prefix_ser}"):
                        self.args[k[len(self.prefix_ser) :]] = v
                    else:
                        self.args[k] = v
            else:
                # Add prefix when merging from another Parset
                for k, v in other.args.items():
                    if not k.startswith("#"):
                        if other.prefix and k == other.prefix:
                            self.set(f"{k}", v)
                        else:
                            self.set(f"{other.prefix}.{k}", v)
                    else:
                        self.comment(v)

    def _deser(self, content):
        config = configparser.ConfigParser(interpolation=None)
        # Preserve case sensitivity
        config.optionxform = str
        # Add a section as there is no section in the file
        content = f"[DEFAULT]\n{content}"
        config.read_string(content)

        result = Parset(prefix=self.prefix)
        for key, v in config["DEFAULT"].items():
            if v == "true":
                v = True
            elif v == "false":
                v = False
            result.set(key, v)
        return result

    def parse_string(self, content):
        defaults = self._deser(content)
        self.merge(defaults, dropPrefix=True)

    def group(self, *group_keys):
        prefix = ".".join(group_keys)

        def group_set(key, value):
            self.set(f"{prefix}.{key}", value)

        return group_set

    def read_from_file(self, filename: str):
        """Read parset configuration from a file

        Args:
            filename: Path to the parset configuration file
        """
        with open(filename) as stream:
            # Add a section as there is no section in the file
            content = stream.read()
            defaults = self._deser(content)
            self.merge(defaults, dropPrefix=True)

    def get(self, key):
        """Get a value by key. If the key is not found and we have a prefix, try with the prefix."""
        try:
            return self.args[key]
        except KeyError:
            if self.prefix:
                prefixed_key = f"{self.prefix_ser}{key}"
                return self.args[prefixed_key]
            raise

    def to_string(self):
        content = []
        for k, v in self.args.items():
            if k == self.prefix:
                content.append(f"{k} = {v}")
            elif not k.startswith("#"):
                if isinstance(v, bool):
                    v = "true" if v is True else "false"
                # print(self.prefix_ser, k, v)
                content.append(f"{self.prefix_ser}{k} = {v}")
            else:
                content.append(k)

        return "\n".join(content) + "\n"

    def serialize(self, filename):
        with open(filename, "w+") as stream:
            stream.write(self.to_string())
            stream.close()



from typing import Union

from taitale.utils.Parset import Parset
from taitale.utils.wrapper import parset_app


def mssplit(
    parset: Union[str, Parset, None] = None,
    **kwargs,
):
    """
    The mssplit program is used to split measurement sets
    """
    if parset is None:
        parset = Parset()
    elif isinstance(parset, str):
        parset_obj = Parset()
        parset_obj.read_from_file(parset)
        parset = parset_obj
    elif not isinstance(parset, Parset):
        raise TypeError("parset must be either a string path or a Parset instance")

    mssplit_app = parset_app(
        name="mssplit",
        parset=parset,
        out_parset_name="taitale_mssplit.in",
        cmd="mssplit -c {parset_name}",
        mpi_compartible=False,
    )

    mssplit_app(**kwargs)


mssplit.__doc__ = """
The mssplit program is used to split measurement sets

Parameters
----------
parset : str or Parset or None, optional
    Either a path to a parset file, a Parset instance, or None to create a new Parset.
    If None, creates a new empty Parset.
**kwargs : dict
    Additional keyword arguments passed to the mssplit application.

Notes
-----
For further information on usage check the ASKAPsoft documentation for `mssplit <https://yandasoft.readthedocs.io/en/latest/calim/mssplit.html>`_

Example
-------

>>> from taitale.askap import mssplit
>>> # Using a parset file
>>> mssplit(
...     parset="mssplit.in",
...     args={
...         "vis": "input.ms",
...         "outputvis": "output.ms",
...         "beams": "[0,1,2]",
...         "channel": "1-300",
...         "width": "1",
...         "column": "DATA",
...         "timebegin": "2010/12/25/00:00:00",
...         "timeend": "2010/12/26/00:00:00",
...         "rows": "0-1000"
...     }
... )

>>> # Using a Parset instance
>>> from taitale.utils.Parset import Parset
>>> parset = Parset()
>>> parset.set("vis", "input.ms")
>>> parset.set("outputvis", "output.ms")
>>> parset.set("beams", "[0,1,2]")
>>> mssplit(parset=parset)
"""



import configparser
import platform
from pathlib import Path, PurePath
from typing import Dict


class TaitaleEnvironment:
    def __init__(self, **kwargs):
        """Initialize TaitaleEnvironment with optional configuration overrides.

        Args:
            runtime : Environment to run ASKAP on 'container' or 'native'.
                Defaults to 'container'
            image : If runtime is 'container', the following image will be used.
                Defaults to  'registry.gitlab.com/askapsdp/all_yandasoft'
            tag : If runtime is 'container', the following image will be used.
                Defaults to '1.1.x-latest-slim'
            engine : If runtime is 'container', the following OCI runtime will be used eg) 'podman or 'docker'.
                Defaults to 'docker'
            workload_manager : Workload manager to use for distribution. Currently supported 'mpi' or 'slurm'.
                Defaults to 'mpi'
            path : Path to executable.
                Defaults to None
            dryrun : If this is set to True, parset is printed to stdout, and execution is not run.
                Defaults to False
            user_defaults : If this is set to True, defaults are additionally read from askap-parsets.cfg.
                Defaults to True
        """
        self.config = {"user_defaults": True}
        self.kwargs = kwargs

    def read_defaults(self, file_name: str) -> configparser.ConfigParser:
        cfg = configparser.ConfigParser()
        cfg.optionxform = str

        cfg.clear()
        package_default_path = PurePath(Path(__file__).parent, file_name)
        if Path(package_default_path).exists():
            with open(package_default_path) as default_file:
                cfg.read_file(default_file)

        if self.config["user_defaults"] is True:
            if platform.system() != "Windows":
                try:
                    user_default_path = PurePath(Path.home(), ".taitale", file_name)
                    if Path(user_default_path).exists():
                        with open(user_default_path) as default_file:
                            cfg.read_file(default_file)
                except Exception:
                    print(
                        f"Error reading user defined config, please check the contents of {user_default_path}"
                    )

            try:
                local_default_path = PurePath(Path("."), file_name)
                if Path(local_default_path).exists():
                    with open(local_default_path) as default_file:
                        cfg.read_file(default_file)
            except Exception:
                print(
                    f"Error reading user defined config, please check the contents of {local_default_path}"
                )

        return cfg

    def yandameta(self) -> Dict:
        file_name = "taitale_env.cfg"
        meta = self.read_defaults(file_name)

        userconfig = dict(meta.items("taitale_env"))

        # Explicitly setting only MPI and Slurm
        userconfig["mpi"] = dict(meta.items("mpi"))
        userconfig["slurm"] = dict(meta.items("slurm"))

        userconfig["native"] = dict(meta.items("native"))
        userconfig["container"] = dict(meta.items("container"))

        return userconfig

    def __enter__(self) -> Dict:
        """Enter the context, setting up the environment configuration."""
        userconfig = self.yandameta()
        self.config = {**userconfig, **self.config, **self.kwargs}

        runtime = self.config.get("runtime")
        if runtime is None or (runtime != "container" and runtime != "native"):
            raise ValueError(
                "Unknown environment, must be either 'container' or 'native'"
            )

        workload_manager = self.config.get("workload_manager")
        if workload_manager is None or (
            workload_manager != "mpi" and workload_manager != "slurm"
        ):
            raise ValueError(
                "Unknown workload_manager, must be either 'mpi' or 'slurm'"
            )

        return self.config

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the context, cleaning up if necessary."""
        pass


# TODO: Revisit this
def taitale_env(**kwargs):
    """
    Args:
        runtime : Environment to run ASKAP on 'container' or 'native'.
            Defaults to 'container'
        image : If runtime is 'container', the following image will be used.
            Defaults to  'registry.gitlab.com/askapsdp/all_yandasoft'
        tag : If runtime is 'container', the following image will be used.
            Defaults to '1.1.x-latest-slim'
        engine : If runtime is 'container', the following OCI runtime will be used eg) 'podman or 'docker'.
            Defaults to 'docker'
        workload_manager : Workload manager to use for distribution. Currently supported 'mpi' or 'slurm'.
            Defaults to 'mpi'
        path : Path to executable.
            Defaults to None
        dryrun : If this is set to True, parset is printed to stdout, and execution is not run.
            Defaults to False
        ipy : If this is set to True, Ipython specific improvements are added.
            Defaults to False
        user_defaults : If this is set to True, defaults are additionally read from askap-parsets.cfg.
            Defaults to True

    Example:
        taitale_env(runtime='container', engine='podman')
        taitale_env(runtime='native', workload_manager='slurm')
    """

    with TaitaleEnvironment(**kwargs) as config:
        return config



from typing import Union

from taitale.utils.Parset import Parset
from taitale.utils.wrapper import parset_app


def imcontsub(
    parset: Union[str, Parset, None] = None,
    **kwargs,
):
    if parset is None:
        parset = Parset(prefix="imcontsub")
    elif isinstance(parset, str):
        parset_obj = Parset(prefix="imcontsub")
        parset_obj.read_from_file(parset)
        parset = parset_obj
    elif not isinstance(parset, Parset):
        raise TypeError("parset must be either a string path or a Parset instance")

    imcontsub_app = parset_app(
        name="imcontsub",
        parset=parset,
        out_parset_name="taitale_imcontsub.in",
        cmd="imcontsub -c {parset_name}",
        mpi_compartible=True,
    )

    imcontsub_app(**kwargs)


imcontsub.__doc__ = """
The purpose of imcontsub is to subtract continuum from an image cube.
The tool processes the cube in parallel using MPI,
where each rank processes a block of Ny/Nranks x-z planes.

Parameters
----------
parset : str or Parset or None, optional
    Either a path to a parset file, a Parset instance, or None to create a new Parset.
    If None, creates a new empty Parset with prefix "imcontsub".
**kwargs : dict
    Additional keyword arguments passed to the imcontsub application.
    Common parameters include:

    - inputfitscube : str
        Image cube to work with. Must be in FITS format.
    - outputfitscube : str
        The name of the output cube. If unspecified, generated from input name
        (e.g., mycube.fits -> mycube.contsub.fits)
    - order : int
        The order of the polynomial used to fit for the continuum (default: 2)
    - threshold : float
        Threshold in robust rms units to decide which channels to include
        in the fit (default: 2.0)
    - blocksize : int
        Size of channel blocks for subtraction. Matches beamforming interval.
        0 means fit/subtract whole spectrum (default: 0)
    - shift : int
        Shift the origin of subtraction blocks left by this many channels (default: 0)
    - interleave : bool
        If true, interleave fit/subtract blocks using blocksize channels but
        subtracting only central 50% (default: false)
    - iterativeclip : bool
        If true, use iterative robust clipping to reject emission/absorption
        features (default: true)
    - imageaccess : str
        IO mode: 'collective' (default, MPI collective IO) or
        'individual' (sequential mode)
    - imageHistory : list[str]
        Lines to add to the image history

Notes
-----
For further information on usage check the ASKAPsoft documentation for `imcontsub <https://yandasoft.readthedocs.io/en/latest/calim/imcontsub.html>`_

Example
-------
>>> from taitale.askap import imcontsub
>>> # Using kwargs directly
>>> imcontsub(
...    workers=2,
...    args={
...        "inputfitscube": "mycube.fits",
...        "outputfitscube": "mycube.contsub.fits",
...        "order": "1",
...        "threshold": "2.5",
...        "blocksize": "54",
...    }
... )
"""



from typing import Union

from taitale.utils.Parset import Parset
from taitale.utils.wrapper import parset_app


def csimulator(
    parset: Union[str, Parset, None] = None,
    **kwargs,
):
    if parset is None:
        parset = Parset(prefix="Csimulator")
    elif isinstance(parset, str):
        parset_obj = Parset(prefix="Csimulator")
        parset_obj.read_from_file(parset)
        parset = parset_obj
    elif not isinstance(parset, Parset):
        raise TypeError("parset must be either a string path or a Parset instance")

    csimulator_app = parset_app(
        name="csimulator",
        parset=parset,
        out_parset_name="taitale_csimulator.in",
        cmd="csimulator -c {parset_name}",
        mpi_compartible=False,
    )

    csimulator_app(**kwargs)


csimulator.__doc__ = """
The csimulator program is used to generate a measurement set from a model sky.

Parameters
----------
parset : str or Parset or None, optional
    Either a path to a parset file, a Parset instance, or None to create a new Parset.
    If None, creates a new empty Parset with prefix "Csimulator".
out_parset_name : str, optional
    The name of the output parset file. Defaults to "taitale_csimulator.in".
**kwargs : dict
    Additional keyword arguments passed to the csimulator application.

Notes
-----
For further information on usage check the ASKAPsoft documentation for `csimulator <https://yandasoft.readthedocs.io/en/latest/calim/csimulator.html>`_

Example
-------
>>> from taitale.askap import csimulator
>>> # Using a parset file
>>> csimulator(
...    parset="csimulator.in",
...     workers=4,
...     args = {
...         "dataset": "test.ms",
...         "antennas": "./parsets/ASKAP36.antpos.in",
...         "feeds": "./parsets/ASKAP1feeds.in",
...         "duration": {
...             "from": "-2h",
...             "to": "2h"
...         },
...         "spw": '[48, 1.420GHz, 1MHz, "XX YY XY YX"]'
...     }
... )

>>> # Using a Parset instance
>>> from taitale.utils.Parset import Parset
>>> parset = Parset(prefix="Csimulator")
>>> parset.set("dataset", "test.ms")
>>> csimulator(parset=parset, workers=4)
"""



import argparse
from dataclasses import dataclass

from .argumentset import ArgumentSet


@dataclass
class CliArgs(ArgumentSet):

    parser: argparse.ArgumentParser = None
    option_sep: str = "--"

    def set(self, key: str, value: any = True, pType="option"):
        if pType == "option":
            self.args[key] = value
        if pType == "positional":
            if "positional" in self.args and isinstance(self.args["positional"], list):
                self.args["positional"].append(key)
            else:
                self.args["positional"] = key

    def _deser(self, content):
        if self.parser is None:
            # raise ValueError("Cannot parse raw arguments without parser")
            return {"positional": content}
        args, _ = self.parser.parse_known_args(content.split(" "))
        return vars(args)

    def parse_string(self, content):
        raw_values = self._deser(content)
        self.merge(ArgumentSet(args=raw_values))

    def to_string(self):
        content = []
        # TODO: Find a better logic for this
        for k, v in self.args.items():
            if k == "positional":
                if isinstance(v, list):
                    content.append(" ".join(v))
                else:
                    content.append(v)
            elif v is True:
                content.append(f"{self.option_sep}{k}")
            elif v is False or v is None:
                pass
            else:
                if isinstance(v, list):
                    content.append(f"{self.option_sep}{k} {' '.join(v)}")
                else:
                    content.append(f"{self.option_sep}{k} {v}")
        return " ".join(content)



from typing import Union

from taitale.utils.Parset import Parset
from taitale.utils.wrapper import parset_app


def dp3(
    parset: Union[str, Parset, None] = None,
    **kwargs,
):
    if parset is None:
        parset = Parset()
    elif isinstance(parset, str):
        parset_obj = Parset()
        parset_obj.read_from_file(parset)
        parset = parset_obj
    elif not isinstance(parset, Parset):
        raise TypeError("parset must be either a string path or a Parset instance")

    dp3_app = parset_app(
        name="DP3",
        parset=parset,
        out_parset_name="taitale_dp3.in",
        cmd="DP3 -c {parset_name}",
        mpi_compartible=False,
    )

    dp3_app(**kwargs)


dp3.__doc__ = """
Taitale interface to DP3

Parameters
----------
parset : str or Parset or None, optional
    Either a path to a parset file, a Parset instance, or None to create a new Parset.
    If None, creates a new empty Parset.
**kwargs : dict
    Additional keyword arguments passed to the DP3 application.

Notes
-----
For further information on usage check the documentation for `DP3 <https://dp3.readthedocs.io/>`_

Example
-------

>>> from taitale.tools import dp3
>>> # Using a parset file
>>> dp3(
...     parset="dp3.in",
...     args={
...         "msin": "MWA-1052736496-averaged.ms",
...         "steps": "[average]",
...         "average.type": "averager",
...         "average.timestep": "8",
...         "msout": "MWA-1052736496-averaged-averaged.MS",
...         "msout.datacolumn": "DATA",
...     },
... )

>>> # Using a Parset instance
>>> from taitale.utils.Parset import Parset
>>> parset = Parset()
>>> parset.set("msin", "MWA-1052736496-averaged.ms")
>>> parset.set("steps", "[]")
>>> parset.set("msout", "MWA-1052736496-averaged-copy.MS")
>>> dp3(parset=parset)
"""
